"""
2d interpolation example
"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.image import NonUniformImage
from sdf.ndtable import NDTable


def peaks(x=np.linspace(-3, 3, 49), y=np.linspace(-3, 3, 49)):
    X, Y = np.meshgrid(x, y)
    Z = 3*(1-X)**2 * np.e**(-(X**2) - (Y+1)**2) - 10*(X/5 - X**3 - Y**5) * np.e**(-X**2-Y**2) - 1/3 * np.e**(-(X+1)**2 - Y**2)
    return X, Y, Z

x = y = np.linspace(-3, 3, 15)
_, _, Z = peaks(x, y)  
table = NDTable(Z, (x, y))

xi = yi = np.linspace(-6, 6, 200)
XI, YI = np.meshgrid(xi, yi, indexing='ij')

figure, axes = plt.subplots(ncols=2, nrows=2, sharex=True, sharey=True)
figure.canvas.set_window_title('Inter- & Extrapolation Methods')
figure.set_facecolor('white')

axes = axes.flatten()

ax = axes[0]
ax.set_title('original')
im = NonUniformImage(ax)
im.set_data(x, y, Z)
im.set_extent((-3, 3, -3, 3))
ax.images.append(im)
ax.set_xlim([-6, 6])
ax.set_ylim([-6, 6])

methods = [('nearest', 'hold'), ('linear', 'linear'), ('akima', 'linear')]

for ax, method in zip(axes[1:], methods):
    ZI = table.evaluate((XI, YI), interp=method[0], extrap=method[1])
    ax.set_title("interp='%s', extrap='%s'" % method)
    im = NonUniformImage(ax)
    im.set_data(xi, yi, ZI)
    im.set_extent((-6, 6, -6, 6))
    ax.images.append(im)

figure.tight_layout()
plt.show()



import sdf
import sys
import matplotlib.pyplot as plt
import numpy as np
import matplotlib.pylab as pylab


def plot_time_series(filename, datasets):

    params = {
        # 'legend.fontsize': 'medium',
        'figure.figsize': (10, 8),
        'axes.labelsize': 'small',
        #   'axes.titlesize': 'medium',
        'xtick.labelsize': 'small',
        'ytick.labelsize': 'small'
    }

    pylab.rcParams.update(params)

    figure, axes = plt.subplots(len(datasets), sharex=True)

    # figure = plt.figure()
    figure.canvas.set_window_title(filename)
    figure.patch.set_facecolor('white')

    for ax, path in zip(axes, datasets):

        dataset = sdf.load(filename, path)

        scale = dataset.scales[0]

        y = dataset.data
        x = scale.data if scale is not None else range(len(y))

        # ax = figure.add_subplot(len(datasets), 1, i + 1)

        ax.plot(x, y, 'b')

        ax.grid(b=True, which='both', color='0.8', linestyle='-')

        ax.set_xlim([np.nanmin(x), np.nanmax(x)])
        ylabel = path
        if dataset.unit is not None:
            ylabel += " / %s" % dataset.unit
        ax.set_ylabel(ylabel)
        ax.margins(y=0.05)

    # x-axis label
    if scale is not None:
        xlabel = scale.name
        if scale.unit is not None:
            xlabel += " / %s" % scale.unit
        else:
            xlabel = 'Index'
        ax.set_xlabel(xlabel)

    figure.tight_layout()

    plt.show()


if __name__ == '__main__':

    try:
        plot_time_series(filename=sys.argv[1], datasets=sys.argv[2:])
    except:
        import traceback
        traceback.print_exc()
        input("Press Enter to continue...")
        exit(1)



# Copyright (c) 2019 Dassault Systemes. All rights reserved.

import numpy as np
from .units import convert_unit
import re
from copy import copy

from . import hdf5

__version__ = '0.3.5'

_object_name_pattern = re.compile('[a-zA-Z][a-zA-Z0-9_]*')


class Group(object):
    """ SDF Group """

    def __init__(self, name, comment=None, attributes=dict(), groups=[], datasets=[]):
        self.name = name
        self.comment = comment
        self.attributes = copy(attributes)
        self.groups = copy(groups)
        self.datasets = copy(datasets)

    def __contains__(self, key):
        for obj in self.datasets + self.groups:
            if obj.name == key:
                return True
        return False

    def __getitem__(self, key):
        for obj in self.datasets + self.groups:
            if obj.name == key:
                return obj
        return None

    def __iter__(self):
        for obj in self.groups + self.datasets:
            yield obj

    def __repr__(self):
        return '<SDF Group "' + self.name + '": [' + ', '.join(map(lambda obj: obj.name, self)) + ']>'


class Dataset(object):
    """ SDF Dataset """

    def __init__(self, name,
                 comment=None,
                 attributes=dict(),
                 data=np.empty(0),
                 display_name=None,
                 relative_quantity=False,
                 unit=None,
                 display_unit=None,
                 is_scale=False,
                 scales=[]
                 ):
        self.name = name
        self.comment = comment
        self.attributes = copy(attributes)
        self.data = data
        self._display_name = display_name
        self.relative_quantity = relative_quantity
        self.unit = unit
        self._display_unit = display_unit
        self.is_scale = is_scale
        self.scales = scales

    @property
    def display_data(self):
        return convert_unit(self.data, self.unit, self.display_unit)

    @display_data.setter
    def display_data(self, value):
        self.data = convert_unit(value, self.display_unit, self.unit)

    @property
    def display_name(self):
        return self._display_name if self._display_name else self.name

    @display_name.setter
    def display_name(self, value):
        self._display_name = value

    @property
    def display_unit(self):
        return self._display_unit if self._display_unit else self.unit

    @display_unit.setter
    def display_unit(self, value):
        self._display_unit = value

    def validate(self):
        if self.display_unit and not self.unit:
            return 'ERROR', 'display_unit was set but no unit'

        return 'OK'

    # some shorthand aliases
    @property
    def d(self):
        return self.data

    dd = display_data

    def __repr__(self):
        text = '<SDF Dataset "' + self.name + '": '

        if not isinstance(self.data, np.ndarray) or len(self.data.shape) == 0:
            text += str(self.data)
        elif len(self.data.shape) == 1 and len(self.data) <= 10:
            text += str(self.data)
        else:
            text += '<' + 'x'.join(map(str, self.data.shape)) + '>'

        if self.unit is not None:
            text += ' ' + self.unit

        if any(self.scales):
            text += ' w.r.t. ' + ', '.join(map(lambda s: s.name if s is not None else 'None', self.scales))

        text += '>'

        return text


def validate(obj):
    """ Validate an sdf.Group or sdf.Dataset """

    errors = []

    if isinstance(obj, Group):
        errors += _validate_group(obj, is_root=True)
    elif isinstance(obj, Dataset):
        errors += _validate_dataset(obj)
    else:
        errors.append('Unknown object type: %s' % type(obj))

    return errors


def _validate_group(group, is_root=False):
    errors = []

    if not is_root and not _object_name_pattern.match(group.name):
        errors += [
            "Object names must only contain letters, digits and underscores (\"_\") and must start with a letter"]

    for child_group in group.groups:
        errors += _validate_dataset(child_group)

    for ds in group.datasets:
        errors += _validate_dataset(ds)

    return errors


def _validate_dataset(ds):

    if not type(ds.data) is np.ndarray:
        return ['Dataset.data must be a numpy.ndarray']

    elif ds.data.size < 1:
        return ['Dataset.data must not be empty']

    elif not np.issubdtype(ds.data.dtype, np.float64):
        return ['Dataset.data.dtype must be numpy.float64']

    if ds.is_scale:
        if len(ds.data.shape) != 1:
            return ['Scales must be one-dimensional']
        if np.any(np.diff(ds.data) <= 0):
            return ['Scales must be strictly monotonic increasing']
    else:
        if (len(ds.data.shape) >= 1) and (ds.data.shape[0] > 0) and not (len(ds.data.shape) == len(ds.scales)):
            return ['The number of scales does not match the number of dimensions']

    return []


def load(filename, objectname='/', unit=None, scale_units=None):
    """ Load a dataset or group from an SDF file """

    obj = None

    if filename.endswith('.mat'):
        from . import dsres
        obj = dsres.load(filename, objectname)
    else:
        obj = hdf5.load(filename, objectname)

    if isinstance(obj, Dataset):

        # check the unit
        if unit is not None and unit != obj.unit:
            raise Exception("Dataset '%s' has the wrong unit. Expected '%s' but was '%s'." % (obj.name, unit, obj.unit))

        # check the number of the scale units
        if scale_units is not None:

            if len(scale_units) != obj.data.ndim:
                raise Exception("The number of scale units must be equal to the number of dimensions. " +
                                "Dataset '%s' has %d dimension(s) but %d scale units where given."
                                % (obj.name, obj.data.ndim, len(scale_units)))

            # check the scale units
            for i, scale_unit in enumerate(scale_units):
                scale = obj.scales[i]
                if scale.unit != scale_unit:
                    raise Exception(("The scale for dimension %d of '%s' has the wrong unit. " +
                                    "Expected '%s' but was '%s'.") % (i + 1, obj.name, scale_unit, scale.unit))

    return obj


def save(filename, group):
    """ Save an SDF group to a file """

    hdf5.save(filename, group)



"""
Create a simple SDF file
"""

import sdf
import numpy as np

# create the data arrays
t = np.linspace(0, 10, 100)
v = np.sin(t)

# create the datasets
ds_t = sdf.Dataset('t', data=t, unit='s', is_scale=True, display_name='Time')
ds_v = sdf.Dataset('v', data=v, unit='V', scales=[ds_t])

# create the root group
g = sdf.Group('/', comment='A sine voltage', datasets=[ds_t, ds_v])

# write the SDF file
sdf.save('sine.sdf', g)

# read the SDF file
ds_v2 = sdf.load('sine.sdf', '/v', unit='V', scale_units=['s'])
ds_t2 = ds_v2.scales[0]

t2 = ds_t2.data
v2 = ds_v2.data



"""
Interpolation Methods
"""

import numpy as np
import matplotlib.pyplot as plt
from sdf.ndtable import NDTable

x = np.linspace(0, 2 * np.pi, 6)
y = np.sin(x)

table = NDTable(y, (x,))

xi = np.linspace(-1.5,  2 * np.pi + 1.5, 1000)
dxi = np.ones_like(xi)

methods = [('hold', 'hold'), ('nearest', 'hold'), ('linear', 'linear'), ('akima', 'linear')]

figure, axes = plt.subplots(len(methods), sharex=True)

figure.canvas.set_window_title('Inter- and Extrapolation Methods')
figure.set_facecolor('white')

for ax, method in zip(axes, methods):
    yi = table.evaluate((xi,), interp=method[0], extrap=method[1])
    dyi = table.evaluate_derivative((xi,), (dxi,), interp=method[0], extrap=method[1])
    ax.set_title("interp='%s', extrap='%s'" % method)
    ax.grid(b=True, which='both', color='0.9', linestyle='-', zorder=0)
    ax.plot(x, y, 'or', label='samples', zorder=300)
    ax.plot(xi, yi, 'b', label='interpolated', zorder=100)
    ax.plot(xi, dyi, 'r', label='derivative', zorder=100)
    ax.set_xlim([-1.5, 2 * np.pi + 1.5])
    ax.set_ylim([-2, 2])

plt.legend(bbox_to_anchor=(0, -0.75, 1., 0.1), loc=8, ncol=3, mode="normal", borderaxespad=0.5)

figure.set_size_inches(w=8, h=8, forward=True)
figure.tight_layout()
figure.subplots_adjust(bottom=0.15, hspace=0.5)
plt.show()



"""
Spline based interpolation methods
"""

import numpy as np
import matplotlib.pyplot as plt
from sdf.ndtable import NDTable


x = np.array([0, 1, 2, 3, 4, 5])
y = np.array([0, 1, 2, 2, 3, 4])

table = NDTable(y, (x,))

xi = np.linspace(-1, 6, 1000)
dxi = np.ones_like(xi)


methods = ['akima', 'fritsch-butland', 'steffen']

figure, axes = plt.subplots(len(methods), sharex=True)

figure.canvas.set_window_title('Spline based interpolation methods')
figure.set_facecolor('white')

for ax, method in zip(axes, methods):
    yi = table.evaluate((xi,), interp=method, extrap='linear')
    dyi = table.evaluate_derivative((xi,), (dxi,), interp=method, extrap='linear')
    ax.set_title(method)
    ax.grid(b=True, which='both', color='0.9', linestyle='-', zorder=0)
    ax.plot(x, y, 'or', label='samples', zorder=300)
    ax.plot(xi, yi, 'b', label='interpolated', zorder=100)
    ax.plot(xi, dyi, 'r', label='derivative', zorder=100)
    ax.margins(x=0, y=0.1)

plt.legend(bbox_to_anchor=(0, -0.75, 1., 0.1), loc=8, ncol=3, mode="normal", borderaxespad=0.5)

figure.tight_layout()
figure.subplots_adjust(bottom=0.18, hspace=0.5)
plt.show()



# Copyright (c) 2019 Dassault Systemes. All rights reserved.

import h5py
import sdf
import numpy as np
import os
import sys


def _to_python_str(s):
    """ Convert to Python string """

    if isinstance(s, bytes):
        return s.decode('utf-8')
    else:
        return s


def load(filename, objectname):

    with h5py.File(filename, 'r') as f:

        datasets = {}

        dsobj = f[objectname]
        class_name = dsobj.__class__.__name__

        if class_name == 'Group':
            group = _create_group(dsobj, datasets)
            _restore_scales(datasets)
            return group
        elif class_name == 'Dataset':
            dataset = _create_dataset(dsobj, datasets)

            for ri in range(dsobj.ndim):
                if dsobj.dims[ri]:
                    sobj = dsobj.dims[ri][0]
                    s = _create_dataset(sobj, dict())
                    s.is_scale = True
                    dataset.scales[ri] = s

            return dataset

        else:
            raise Exception('Unexpected object')


def save(filename, group):

    with h5py.File(filename, 'w') as f:

        datasets = dict()
        _write_group(f, group, '/', datasets)

        # attach the scales
        for ds, h5ds in datasets.items():
            for i, s in enumerate(ds.scales):
                if s is None:
                    continue
                elif s in datasets:
                    h5s = datasets[s]
                    dimname = s._display_name
                    if dimname is None:
                        dimname = ''
                    h5s.make_scale(_str(dimname))
                    h5ds.dims[i].attach_scale(h5s)
                else:
                    print("Cannot attach scale for '" + h5ds.name +
                          "' because the referenced scale for dimension " + str(i) + " is not part of the file")


def _create_group(gobj, datasets):
    """ Create an sdf.Group from an h5py group """

    ds_obj_list = []
    g_obj_list = []

    group_attrs = {key: gobj.attrs[key] for key in gobj.attrs.keys() if key != 'COMMENT'}
    comment = gobj.attrs.get('COMMENT')

    for ds_name in gobj.keys():
        # TODO: fix this?
        if isinstance(gobj[ds_name], h5py._hl.dataset.Dataset):
            ds_obj_list.append(gobj[ds_name])
        elif isinstance(gobj[ds_name], h5py._hl.group.Group):
            g_obj_list.append(gobj[ds_name])

    child_groups = []

    for cgobj in g_obj_list:
        child_groups.append(_create_group(cgobj, datasets))

    ds_list = [_create_dataset(dsobj, datasets) for dsobj in ds_obj_list]

    name = gobj.name.split('/')[-1]

    return sdf.Group(name=name, comment=comment, attributes=group_attrs, groups=child_groups, datasets=ds_list)


def _create_dataset(dsobj, datasets):
    """ Create a dataset from an h5py dataset """

    _, name = os.path.split(dsobj.name)
    ds = sdf.Dataset(name, data=dsobj[()])

    for attr in dsobj.attrs:
        if attr == 'COMMENT':
            ds.comment = _to_python_str(dsobj.attrs[attr])
        elif attr == 'NAME':
            ds.display_name = _to_python_str(dsobj.attrs[attr])
        elif attr == 'RELATIVE_QUANTITY' and _to_python_str(dsobj.attrs[attr]) == 'TRUE':
            ds.relative_quantity = True
        elif attr == 'UNIT':
            ds.unit = _to_python_str(dsobj.attrs[attr])
        elif attr == 'DISPLAY_UNIT':
            ds.display_unit = _to_python_str(dsobj.attrs[attr])
        elif attr == 'CLASS' and _to_python_str(dsobj.attrs[attr]) == 'DIMENSION_SCALE':
            ds.is_scale = True
        elif attr == 'REFERENCE_LIST':
            ds.is_scale = True
        elif attr in ['REFERENCE_LIST', 'DIMENSION_LIST']:
            pass
        else:
            ds.attributes[attr] = _to_python_str(dsobj.attrs[attr])

    ds.scales = [None] * ds.data.ndim

    datasets[dsobj] = ds

    return ds


def _restore_scales(datasets):

    for dsobj, ds in datasets.items():

        for i in range(ds.data.ndim):
            if dsobj.dims[i]:
                sobj = dsobj.dims[i][0]
                scale = datasets[sobj]
                scale.is_scale = True
                ds.scales[i] = scale
                pass


def _str(s):
    """ Convert to byte string """

    if sys.version_info.major >= 3 and isinstance(s, bytes):
        return s
    else:
        # convert the string to an fixed-length utf-8 byte string
        return np.string_(s.encode('utf-8'))


def _write_group(f, g, path, datasets):

    if path == '/':
        gobj = f
    else:
        gobj = f.create_group(path)

    # iterate over the child groups
    for subgroup in g.groups:
        _write_group(f, subgroup, path + subgroup.name + '/', datasets)

    if g.comment is not None:
        gobj.attrs['COMMENT'] = _str(g.comment)

    for key, value in g.attributes.items():
        gobj.attrs[key] = _str(value)

    # write the datasets
    for ds in g.datasets:
        _write_dataset(f, ds, path, datasets)


def _write_dataset(f, ds, path, datasets):

    f[path + ds.name] = ds.data
    dsobj = f[path + ds.name]

    datasets[ds] = dsobj

    if ds.comment:
        dsobj.attrs['COMMENT'] = _str(ds.comment)

    if ds._display_name:
        dsobj.attrs['NAME'] = _str(ds.display_name)

    if ds.relative_quantity:
        dsobj.attrs['RELATIVE_QUANTITY'] = _str('TRUE')

    if ds.unit:
        dsobj.attrs['UNIT'] = _str(ds.unit)

    if ds.display_unit != ds.unit:
        dsobj.attrs['DISPLAY_UNIT'] = _str(ds.display_unit)

    if ds.is_scale:
        dimname = ds.display_name

        if dimname is None:
            dimname = ''

        h5py.h5ds.set_scale(dsobj.id, _str(dimname))

    return dsobj



"""
Create a contour plot from an SDF dataset
"""

import numpy as np
import sdf
import sys
import matplotlib.pyplot as plt


filename = sys.argv[1]
objects = sys.argv[2]

dataset = sdf.load(filename, objects)

delta = 0.025


x = dataset.scales[1].data
y = dataset.scales[0].data

subs = [0] * dataset.data.ndim
subs[0] = slice(None)
subs[1] = slice(None)

Z = dataset.data[subs].T

X, Y = np.meshgrid(x, y, indexing='ij')

figure = plt.figure()

figure.patch.set_facecolor('white')

ax = figure.add_subplot(1, 1, 1)

ax.grid(True)

CS = plt.contourf(X, Y, Z, 10, cmap=plt.cm.viridis)

cbar = figure.colorbar(CS)

CS = plt.contour(X, Y, Z, 10, colors='k')

plt.clabel(CS=CS,  fontsize=9, inline=1, colors='k')

ax.set_title(dataset.display_name + " / " + dataset.unit)
ax.set_xlabel(dataset.scales[1].display_name + " / " + dataset.scales[1].unit)
ax.set_ylabel(dataset.scales[0].display_name + " / " + dataset.scales[0].unit)

plt.tight_layout()

plt.show()



# Copyright (c) 2019 Dassault Systemes. All rights reserved.

import math


class Table(object):
    """
    Utility class to store the unit conversions (inspired by Guava's com.google.common.collect.Table<R,C,V>)
    """

    def __init__(self):
        self._rows = dict()
        
    def put(self, row_key, column_key, value):
        if row_key in self._rows:
            self._rows[row_key][column_key] = value
        else:
            self._rows[row_key] = {column_key: value}
            
    def get(self, row_key, column_key):
        if row_key in self._rows:
            return self._rows[row_key][column_key]
        else:
            return None


global _converters
_converters = Table()


def convert_unit(value, from_unit, to_unit):
    
    if from_unit == to_unit:
        return value # nothing to do
    
    converter = _converters.get(from_unit, to_unit)
    
    if not converter:
        raise Exception('No conversion defined for "' + from_unit + '" -> "' + to_unit + '"')
    
    return converter.convert(value)


class LinearUnitConverter(object):
    
    def __init__(self, from_unit, to_unit, factor, offset=0):
        self.from_unit = from_unit
        self.to_unit = to_unit
        self.factor = factor
        self.offset = offset
        
    def convert(self, value):
        return value * self.factor + self.offset


def define_unit_conversion(from_unit, to_unit, factor, offset=0):
    global _converters
    _converters.put(from_unit, to_unit, LinearUnitConverter(from_unit, to_unit, factor, offset))
    _converters.put(to_unit, from_unit, LinearUnitConverter(to_unit, from_unit, 1/factor, -offset/factor))
    
# Time
define_unit_conversion('s', 'ms', 1000)
define_unit_conversion('s', 'min', 1.0/60.0)
define_unit_conversion('s', 'h', 1.0/3600.0)
define_unit_conversion('s', 'd', 1.0/86400.0)

# Angle
define_unit_conversion('rad', 'deg', 180/math.pi)

# Angular velocity
define_unit_conversion('rad/s', 'deg/s', 180/math.pi)
define_unit_conversion('rad/s', 'rpm', 30/math.pi)
define_unit_conversion('rad/s', '1/min', 30/math.pi)
define_unit_conversion('rad/s', 'r/min', 30/math.pi)

# Length, distance
define_unit_conversion('m', 'km', 0.001)
# define_unit_conversion('m', 'cm', 100)
define_unit_conversion('m', 'mm', 1000)

# Area
# define_unit_conversion('m2','cm2',1e4)

# Volume
define_unit_conversion('m3', 'l', 1e3)
define_unit_conversion('m3', 'ml', 1e6)

# Pressure
define_unit_conversion('Pa', 'kPa', 1e-3)
define_unit_conversion('Pa', 'MPa', 1e-6)
define_unit_conversion('Pa', 'bar', 1e-5)

# Bulk Modulus
define_unit_conversion('N/m2','bar',1/1e5)

# Volume Flow Rate
define_unit_conversion('m3/s', 'l/min', 1e3*60)

# Density
define_unit_conversion('kg/m3', 'kg/dm3', 1e-3)
define_unit_conversion('kg/m3', 'kg/l', 1e-3)
# define_unit_conversion('kg/m3', 'g/cm3', 1e-3)
# define_unit_conversion('kg/s', 'g/s', 1e3)

# Speed
define_unit_conversion('m/s', 'km/h', 3.6)
define_unit_conversion('m/s', 'mm/s', 1e3)
# define_unit_conversion('m/s', 'knots', 1.9438445)

# Force
define_unit_conversion('N', 'mN', 1000)
define_unit_conversion('N', 'kN', 1e-3)
define_unit_conversion('N', 'MN', 1e-6)

# Work, Energy
define_unit_conversion('J', 'kWh', 1.0/3600.0/1000.0)
define_unit_conversion('J', 'mJ', 1000)
define_unit_conversion('J', 'kJ', 1e-3)
define_unit_conversion('J', 'MJ', 1e-6)

# Power
define_unit_conversion('W', 'mW', 1000)
define_unit_conversion('W', 'kW', 1e-3)
define_unit_conversion('W', 'MW', 1e-6)

# Temperature
define_unit_conversion('K', 'degC', 1, -273.15)

# Voltage
define_unit_conversion('V', 'mV', 1000)
define_unit_conversion('V', 'kV', 0.001)

# Currrent
define_unit_conversion('A', 'mA', 1000)
define_unit_conversion('A', 'kA', 0.001)

# Capacitance
define_unit_conversion('F', 'uF', 1e6)

# Leakage
define_unit_conversion('m3/(s.Pa)','l/(min.bar)', 6e9)

# Viscous Friction
#define_unit_conversion('N.m/(rad/s)', 'N.m/(rev/min)',0.10471975512)

# Kinematic Viscosity
define_unit_conversion('m2/s', 'mm2/s', 1e6)

# Temperature coefficient (e.g. of resistance)
define_unit_conversion('1/K', 'ppm/K', 1e6)



# Copyright (c) 2019 Dassault Systemes. All rights reserved.

import numpy as np
from sdf import Group, Dataset
import scipy.io

# extract strings from the matrix
strMatNormal = lambda a: [''.join(s).rstrip() for s in a]
strMatTrans = lambda a: [''.join(s).rstrip() for s in zip(*a)]


def _split_description(comment):

    unit = None
    display_unit = None
    info = dict()

    if comment.endswith(']'):
        i = comment.rfind('[')
        unit = comment[i + 1:-1]
        comment = comment[0:i].strip()

    if unit is not None:

        if ':#' in unit:
            segments = unit.split(':#')
            unit = segments[0]
            for segment in segments[1:]:
                key, value = segment[1:-1].split('=')
                info[key] = value

        if '|' in unit:
            unit, display_unit = unit.split('|')

    return unit, display_unit, comment, info


def load(filename, objectname):

    g_root = _load_mat(filename)

    if objectname == '/':
        return g_root
    else:
        obj = g_root
        segments = objectname.split('/')
        for s in segments:
            if s:
                obj = obj[s]
        return obj


def _load_mat(filename):

    mat = scipy.io.loadmat(filename, chars_as_strings=False)

    _vars = {}
    _blocks = []

    try:
        fileInfo = strMatNormal(mat['Aclass'])
    except KeyError:
        raise Exception('File structure not supported!')

    if fileInfo[1] == '1.1':
        if fileInfo[3] == 'binTrans':
            # usually files from OpenModelica or Dymola auto saved,
            # all methods rely on this structure since this was the only
            # one understand by earlier versions
            names = strMatTrans(mat['name'])  # names
            descr = strMatTrans(mat['description'])  # descriptions

            cons = mat['data_1']
            traj = mat['data_2']

            d = mat['dataInfo'][0, :]
            x = mat['dataInfo'][1, :]

        elif fileInfo[3] == 'binNormal':
            # usually files from dymola, save as...,
            # variables are mapped to the structure above ('binTrans')
            names = strMatNormal(mat['name'])  # names
            descr = strMatNormal(mat['description'])  # descriptions

            cons = mat['data_1'].T
            traj = mat['data_2'].T

            d = mat['dataInfo'][:, 0]
            x = mat['dataInfo'][:, 1]
        else:
            raise Exception('File structure not supported!')

        c = np.abs(x) - 1  # column
        s = np.sign(x)  # sign

        vars = zip(names, descr, d, c, s)
    elif fileInfo[1] == '1.0':
        # files generated with dymola, save as..., only plotted ...
        # fake the structure of a 1.1 transposed file
        names = strMatNormal(mat['names'])  # names
        _blocks.append(0)
        mat['data_0'] = mat['data'].transpose()
        del mat['data']
        _absc = (names[0], '')
        for i in range(1, len(names)):
            _vars[names[i]] = ('', 0, i, 1)
    else:
        raise Exception('File structure not supported!')

    # build the SDF tree
    g_root = Group('/')

    ds_time = None

    for name, desc, d, c, s in vars:

        unit, display_unit, comment, info = _split_description(desc)

        path = name.split('.')

        g_parent = g_root

        for segment in path[:-1]:
            if segment in g_parent:
                g_parent = g_parent[segment]
            else:
                g_child = Group(segment)
                g_parent.groups.append(g_child)
                g_parent = g_child
            pass

        if d == 1:
            data = cons[c, 0] * s
        else:
            data = traj[c, :] * s

        if 'type' in info:
            if info['type'] == 'Integer' or 'Boolean':
                data = np.asarray(data, dtype=np.int32)

        if d == 0:
            ds = Dataset(path[-1], comment="Simulation time", unit=unit, display_unit=display_unit, data=data)
            ds_time = ds
        elif d == 1:
            ds = Dataset(path[-1], comment=comment, unit=unit, display_unit=display_unit, data=data)
        else:
            ds = Dataset(path[-1], comment=comment, unit=unit, display_unit=display_unit, data=data, scales=[ds_time])

        g_parent.datasets.append(ds)

    return g_root



# Copyright (c) 2019 Dassault Systemes. All rights reserved.

from ctypes import c_double, c_void_p, c_int, byref, cdll
import numpy as np
import sys
import os


_arch = '64' if sys.maxsize > 2 ** 32 else '32'
_dll_path = os.path.join(os.path.abspath(os.path.dirname(__file__)))

if sys.platform.startswith('darwin'):
    _dll_path = os.path.join(_dll_path, 'darwin64', 'libNDTable.dylib')
elif sys.platform.startswith('win'):
    _dll_path = os.path.join(_dll_path, 'win' + _arch, 'ndtable.dll')
elif sys.platform.startswith('linux'):
    _dll_path = os.path.join(_dll_path, 'linux' + _arch, 'libndtable.so')
else:
    raise Exception("Unsupported platform: " + sys.platform)

_ndtable = cdll.LoadLibrary(_dll_path)

# PYTHON_API ModelicaNDTable_h create_table(int ndims, const int *dims, const double *data, const double **scales) {
_create_table = _ndtable.create_table
_create_table.argtypes = [c_int, c_void_p, c_void_p, (c_void_p * 32)]
_create_table.restype = c_void_p

# PYTHON_API int evaluate(
#     ModelicaNDTable_h table, 
#     int ndims, 
#     const double **params, 
#     ModelicaNDTable_InterpMethod_t interp_method,
#     ModelicaNDTable_ExtrapMethod_t extrap_method,
#     int nvalues,
#     double *value);
_evaluate = _ndtable.evaluate
_evaluate.argtypes = [c_void_p, c_int, c_void_p, c_int, c_int, c_int, c_void_p]
_evaluate.restype = c_int

# PYTHON_API int evaluate_derivative(
#     ModelicaNDTable_h table, 
#     int nparams, 
#     const double params[],
#     const double delta_params[],
#     ModelicaNDTable_InterpMethod_t interp_method,
#     ModelicaNDTable_ExtrapMethod_t extrap_method, 
#     double *value);
_evaluate_derivative = _ndtable.evaluate_derivative
_evaluate_derivative.argtypes = [c_void_p, c_int, c_void_p, c_void_p, c_int, c_int, c_void_p]
_evaluate_derivative.restype = c_int

_close_table = _ndtable.close_table
_close_table.argtypes = [c_void_p]


class NDTable(object):
    """ 
        An n-dimensional lookup table
    
        Attributes
        ----------
        data : ndarray
            The values to interpolate.
        scales : tuple of ndarrays
            The scales for `data`. There must be one scale for every dimension of `data`. 
            The values must be strictly monotonic increasing.
    """
    
    _interp_methods = {'hold': 1, 'nearest': 2, 'linear': 3, 'akima': 4, 'fritsch-butland': 5, 'steffen': 6}
    _extrap_methods = {'hold': 1, 'linear': 2}
    
    def __init__(self, data, scales):

        scales = list(scales)

        # convert the arguments to double arrays
        data = np.asanyarray(data, dtype=np.float64, order='C')

        for i, scale in enumerate(scales):
            scales[i] = np.asanyarray(scale, dtype=np.float64, order='C')

        # check the arguments
        assert data.ndim <= 32, 'Max. number of dimensions is 32'
        assert len(scales) == data.ndim, 'The number of scales must match the number of dimensions'
        for i, scale in enumerate(scales):
            assert np.all(np.isfinite(scale)), 'The scale for dimension %d is not finite' % i
            assert scale.ndim == 1, 'Scales must be one-dimensional'
            assert scale.size == data.shape[i], 'The scale for dimension %d does not match the shape of data' % i
        
        dims = np.asarray(data.shape, np.int32)
        scales_ = (c_void_p * 32)()
        for i, scale in enumerate(scales):
            scales_[i] = scale.ctypes.get_as_parameter()
        self._table = _create_table(c_int(data.ndim), dims.ctypes.get_as_parameter(), data.ctypes.get_as_parameter(), scales_)
        
        # save close function from garbage collection
        self._close_table = _close_table

    def evaluate(self, points, interp='linear', extrap='hold'):
        """
            Evaluate the lookup table at the coordinates in `points`.
            
            Returns an array of the same shape as the coordinates in `points`.
            
            Parameters
            ----------
            points : tuple of ndarrays
                The coordinates of the points to evaluate.
            interp : string, optional
                The interpolation method (one of 'nearest', 'linear' or 'akima')
                Default is 'linear'.
            extrap : string, optional
                The extrapolation method (one of 'hold' or 'linear')
                Default is 'hold'.
        
            Returns
            -------
            samples : ndarray
                The evaluated points.
        
            Example
            --------
            >>> import numpy as np
            >>> from sdf.ndtable import NDTable
            >>> x = y = np.linspace(-1, 1, 3)
            >>> x, y
                (array([-1.,  0.,  1.]), array([-1.,  0.,  1.]))
            >>> X, Y = np.meshgrid(x, y, indexing='ij')
            >>> Z = X * Y
            >>> Z
                array([[ 1., -0., -1.],
                       [-0.,  0.,  0.],
                       [-1.,  0.,  1.]])
            >>> lut = NDTable(Z, (x, y))
            >>> x2 = y2 = np.linspace(-1, 1, 5)
            >>> x2, y2
                (array([-1. , -0.5,  0. ,  0.5,  1. ]), array([-1. , -0.5,  0. ,  0.5,  1. ]))
            >>> X2, Y2 = np.meshgrid(x2, y2, indexing='ij')
            >>> Z2 = lut.evaluate((X2, Y2))
            >>> Z2
                array([[ 1.  ,  0.5 ,  0.  , -0.5 , -1.  ],
                       [ 0.5 ,  0.25,  0.  , -0.25, -0.5 ],
                       [ 0.  ,  0.  ,  0.  ,  0.  ,  0.  ],
                       [-0.5 , -0.25,  0.  ,  0.25,  0.5 ],
                       [-1.  , -0.5 ,  0.  ,  0.5 ,  1.  ]])
        
        """

        points = list(points)
        
        for i, _ in enumerate(points):
            points[i] = np.asarray(points[i], np.float64)

        shape = points[0].shape

        for p in points[1:]:
            assert p.shape == shape, 'The arrays in points must have the same shape'
        
        assert interp in self._interp_methods, 'Unknown interpolation method "%s"' % interp
        assert extrap in self._extrap_methods, 'Unknown extrapolation method "%s"' % extrap
        
        interp_method = c_int(self._interp_methods[interp])
        extrap_method = c_int(self._extrap_methods[extrap])
        
        values = np.empty(shape)
        params = (c_void_p * len(points))()
        for i, param in enumerate(points):
            params[i] = param.ctypes.get_as_parameter()
        
        ret = _evaluate(c_void_p(self._table),
                        c_int(len(params)),
                        params, 
                        interp_method, 
                        extrap_method,
                        c_int(values.size),
                        values.ctypes.get_as_parameter())
        
        assert ret == 0, 'An error occurred during interpolation'
        
        return values

    def evaluate_derivative(self, points, deltas, interp='linear', extrap='hold'):

        points = list(points)
        deltas = list(deltas)
        
        for i, _ in enumerate(points):
            points[i] = np.asarray(points[i], np.float64)
            deltas[i] = np.asarray(deltas[i], np.float64)

        shape = points[0].shape

        for p in points[1:]:
            assert p.shape == shape, 'The arrays in points must have the same shape'
        
        assert interp in self._interp_methods, 'Unknown interpolation method "%s"' % interp
        assert extrap in self._extrap_methods, 'Unknown extrapolation method "%s"' % extrap
        
        interp_method = c_int(self._interp_methods[interp])
        extrap_method = c_int(self._extrap_methods[extrap])
        value = c_double()
        
        out = np.empty(shape)
        params = np.empty(len(points))
        delta_params = np.empty(len(points))
        
        for index in np.ndindex(shape):
            for i, point in enumerate(points):
                params[i] = point[index]
            for i, delta in enumerate(deltas):
                delta_params[i] = delta[index]
                
            _evaluate_derivative(c_void_p(self._table),
                                 c_int(params.size), 
                                 params.ctypes.get_as_parameter(), 
                                 delta_params.ctypes.get_as_parameter(),
                                 interp_method, 
                                 extrap_method, 
                                 byref(value))
            out[index] = value.value
            
        return out

    def __del__(self):
        self._close_table(self._table)



"""
Create a contour plot from an SDF dataset
"""
import traceback

import numpy as np
import sdf
import matplotlib.pyplot as plt
from matplotlib import colors
import sys

import matplotlib.pylab as pylab


def create_plot(filename, datasets):

    params = {
        # 'legend.fontsize': 'medium',
        # 'figure.figsize': (10, 8),
        'legend.fontsize': 'small',
        'axes.labelsize': 'small',
        'axes.titlesize': 'small',
        'xtick.labelsize': 'small',
        'ytick.labelsize': 'small'
    }

    pylab.rcParams.update(params)

    figure = plt.figure(figsize=(12, 8))
    figure.canvas.set_window_title(filename)
    figure.patch.set_facecolor('white')

    nrows = len(datasets)

    for row, dataset in enumerate(datasets):

        # load the datasets
        C1 = sdf.load(filename, dataset)

        ncols = C1.scales[2].data.size

        norm = colors.Normalize(vmin=np.nanmin(C1.data), vmax=np.nanmax(C1.data))

        subs = [0] * C1.data.ndim
        subs[0] = slice(None)
        subs[1] = slice(None)

        ax0 = None

        for i in range(ncols):

            x = C1.scales[1].data
            y = C1.scales[0].data

            subs[2] = i

            Z = C1.data[subs].T

            X, Y = np.meshgrid(x, y, indexing='ij')

            ax = figure.add_subplot(len(datasets), C1.data.shape[2], (row * ncols) + i + 1, sharex=ax0, sharey=ax0)

            if i == 0:
                ax0 = ax
                ax.set_ylabel(C1.display_name + " / " + C1.unit + "\n" + C1.scales[0].display_name + " / " + C1.scales[0].unit)
            # else:
            #     ax.get_yaxis().set_ticklabels([])

            ax.grid(True)

            CSF = plt.contourf(X, Y, Z, 10, cmap=plt.cm.viridis, norm=norm)

            CS = plt.contour(X, Y, Z, 10, colors='k')

            plt.clabel(CS=CS, fontsize=9, inline=1, colors='k')

            scale3 = C1.scales[2]

            ax.set_title(scale3.display_name + "=" + ("%g" % scale3.data[i]) + " " + scale3.unit)
            ax.set_xlabel(C1.scales[1].display_name + " / " + C1.scales[1].unit)

        cbar = figure.colorbar(CSF)

    plt.tight_layout()

    plt.show()

if __name__ == '__main__':

    try:
        create_plot(filename=sys.argv[1], datasets=sys.argv[2:])
    except:
        traceback.print_exc()
        input("Press Enter to continue...")
        exit(1)



"""
Import data from an Excel sheet to SDF
"""

import os.path
import xlrd
from numpy import array
import sdf


# name of the Excel file to import
filename = 'time_series.xlsx'

# open the workbook
book = xlrd.open_workbook(filename)
    
# get the first sheet
sh = book.sheet_by_index(0)

# get the names, quantities and units
n_t = sh.cell_value(0, 1)
u_t = sh.cell_value(1, 1)

n_u = sh.cell_value(0, 2)
u_u = sh.cell_value(1, 2)

# get the data
col_t = sh.col_values(1, 2, sh.nrows)
col_u = sh.col_values(2, 2, sh.nrows)

# create the data arrays
t = array(col_t)
u = array(col_u)

# create the datasets
ds_t = sdf.Dataset(n_t, data=t, unit=u_t, is_scale=True, display_name='Time')
ds_u = sdf.Dataset(n_u, data=u, unit=u_u, scales=[ds_t])

# create the root group
g = sdf.Group('/', comment='Imported from ' + filename, datasets=[ds_t, ds_u])

# change the file extension
outfile = os.path.splitext(filename)[0] + '.sdf'

# write the SDF file
sdf.save(outfile, g)
